# This script uses opencv to observe key-presses in piano-tutorial videos
# and converts them into a MIDI file
# feel free to use, edit & share
#
# To create your own layout run keyboard_editor.py and create a new profile
#
# feel free to use, edit & share
# Written by gilex, 2023 (https://github.com/gilex-dev)
import os
import cv2
import json
from midiutil import MIDIFile

DEBUG = False
frame_limit = -1  # -1 for whole source file
with_color = False  # don't use default color of key

# USE BGR and NOT RGB
if with_color:
    tolerance = 10
    colors = {"whiteKey": {"blue_light": [255, 255, 163], "green_light": [113, 255, 252]},
              "blackKey": {"blue_dark": [255, 132, 50], "green_dark": [29, 255, 100], }}
else:
    tolerance = 40
    colors = {"whiteKey": [255, 255, 255], "blackKey": [17, 8, 16]}

source_video_file = 'video_trim.mp4'

# User defined variables
cap = cv2.VideoCapture(source_video_file)
if not cap.isOpened():
    print(f"Could not open : {source_video_file}")
    exit()
length = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
if 0 < frame_limit < length:
    length = frame_limit
width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
fps = cap.get(cv2.CAP_PROP_FPS)

timeline = {0: {}}  # buffer for timeline
pianoKeys = dict()


def load_profile():
    count = 0
    profiles = os.listdir('{0}/profiles/'.format(os.getcwd().replace("\\", "/")))
    print("Select a profile")
    for profile in profiles:
        count += 1
        print(f'{count}. {profile.replace(".json", "")}')
    if count > 1:
        count = int(input("Enter a number: ")) - 1
    else:
        count = 0
    with open(f"profiles/{profiles[count]}", 'r') as input_file:
        return json.loads(input_file.read())


def getTimeline(cap):
    print(
        f"{'-' * 20}\n Analyzing {length} frames of {source_video_file}\n tolerance = +/-{round(tolerance / 2.55, 2)}% \n DEBUG = {DEBUG}\n with_color = {with_color}\n{'-' * 20}")
    count = 0  # frame count
    prev_progress = 0
    success, image = cap.read()
    for pianoKey in pianoKeys:
        pianoKeys[pianoKey].update({"lastCount": 0, "keyDown": 0})
    while success:
        count += 1
        progress = int(count / length * 100)
        if progress > prev_progress:
            prev_progress += 10
            print(f"Processing... {progress}%")
        for pianoKey in pianoKeys:
            v = image[pianoKeys[pianoKey]["h"], pianoKeys[pianoKey]["w"]]  # numpy array
            # if DEBUG: (f"{pianoKey}: {v} at frame {count}")
            # if not len(pianoKeys[pianoKey]["default"]): # not working anymore -> used to set color for keys with missing value
            #     pianoKeys[pianoKey]["default"] = v
            if with_color:
                matched_channel = False
                is_key = False
                for color in colors[pianoKeys[pianoKey]["default"]]:  # iterate trough colors
                    highlight_color = colors[pianoKeys[pianoKey]["default"]][color]
                    for bgr in range(3):  # iterate trough (B, G, R)
                        if highlight_color[bgr] - tolerance < v[bgr] < highlight_color[
                            bgr] + tolerance:
                            # print(f'{i}: {pianoKeys[pianoKey]["default"][i] - tolerance} < {v[i]} < {pianoKeys[pianoKey]["default"][i] + tolerance}')
                            matched_channel = True
                        else:
                            matched_channel = False
                            break
                    if matched_channel:
                        is_key = matched_channel
                        break
            else:
                is_key = False
                for i in range(3):
                    if not colors[pianoKeys[pianoKey]["default"]][i] - tolerance < v[i] < \
                           colors[pianoKeys[pianoKey]["default"]][i] + tolerance:
                        is_key = True
            if is_key:  # if pianoKey is not default color
                # print(f"{pianoKey}: {v} at frame {count}")
                if not pianoKeys[pianoKey]["lastCount"] == count - 1:  # new keypress
                    pianoKeys[pianoKey].update({"keyDown": count})
                    if not count in timeline:
                        timeline.update({count: {pianoKey: 1}})
                    else:
                        timeline[count].update({pianoKey: 1})
                else:  # 'middle-frame'
                    timeline[pianoKeys[pianoKey]["keyDown"]].update({pianoKey: count - pianoKeys[pianoKey][
                        "keyDown"]})  # increment duration by 1 -> This is a VERY lazy and not optimized implementation by me.
                pianoKeys[pianoKey].update({"lastCount": count})
            elif pianoKeys[pianoKey]["lastCount"] == count - 1 and DEBUG:
                print(f'{pianoKey} down from {pianoKeys[pianoKey]["keyDown"]} to {count - 1}')
        if count == frame_limit:
            break

        success, image = cap.read()
    cap.release()
    cv2.destroyAllWindows()
    if DEBUG: print(timeline)


def write_midi(beats, output_path, source_fps, title="vis2MIDI"):
    # write to MIDI

    # create your MIDI object
    mf = MIDIFile(1)  # only 1 track
    track = 0  # the only track

    time = 0  # start at the beginning
    mf.addTrackName(track, time, title)
    mf.addTempo(track, time, 60 * 30 / source_fps)
    for frame in beats:
        try:
            for note in beats[frame]:
                channel = 0
                time = int(frame) / source_fps
                duration = int(beats[frame][note]) / source_fps
                pitch = pianoKeys[note]["pitch"]
                volume = 100
                mf.addNote(track, channel, pitch, time, duration, volume)
        except KeyError:
            print(KeyError)

    # write it to disk
    with open(output_path, 'wb') as output_file:
        mf.writeFile(output_file)


def import_from_text(input_path):
    with open(input_path, 'r') as input_file:
        return json.loads(input_file.read())


def save_to_text(output_path, timeline_out):
    with open(output_path, 'w') as output_file:
        output_file.write(json.dumps(timeline_out))


if __name__ == "__main__":
    pianoKeys = load_profile()
    getTimeline(cap)
    save_to_text(f'{source_video_file}.json', timeline)
    timeline = import_from_text(f'{source_video_file}.json')
    write_midi(timeline, f'output/{source_video_file}.mid', fps, source_video_file)
